#include <shmport/ShmBuffer_primitive.hpp>


////////////////////////////////////////////////////////////////////////////////
/*!
 *  @brief コンストラクタ
 *  コンストラクタ
 *  @param shmSegment このクラスのインスタンスを置く共有メモリ領域
 *  @param maxBufferSize バッファに入る型Tの要素の総数
 *
 *  @pre 呼び出し元がshmSegmentの構築，解体を担当する。ShmBufferのインスタンスが
 *       存在する間，shmSegmentは維持される必要がある。
 */
template<typename T>
ShmBuffer<T>::ShmBuffer(boost::interprocess::managed_shared_memory* shmSegment,
		    size_t maxBufferSize) :
  m_shmSegment(shmSegment),
  m_allocator(m_shmSegment->get_segment_manager()),
  m_maxBufferSize(maxBufferSize)
{
  using namespace boost::interprocess;
  m_buffOffset = m_shmSegment->construct<ShmDeque>(anonymous_instance)(m_allocator);
}

////////////////////////////////////////////////////////////////////////////////
/*!
 * @brief デストラクタ
 * デストラクタ。コンストラクタで渡されたshmSegmentは解体しない。
 */
template<typename T>
ShmBuffer<T>::~ShmBuffer()
{
  m_shmSegment->destroy_ptr(m_buffOffset.get());
}

////////////////////////////////////////////////////////////////////////////////
/*!
 *  @brief コピーコンストラクタ
 *  ShmBufferはコピーして使えるようには設計していないのでprivateにする。
 */
template<typename T>
ShmBuffer<T>::ShmBuffer(const ShmBuffer<T>& shmBuffer) :
  m_shmSegment(shmBuffer.m_shmSegment),
  m_allocator(m_shmSegment->get_segment_manager())
{}


////////////////////////////////////////////////////////////////////////////////
/*!
 * @brief バッファの末尾に要素を追加する。
 *  バッファの末尾に要素を追加する。追加により要素数がコンストラクタで指定した
 *  要素数を越えないように、余剰の要素は先頭から取り除かれる。
 *
 */
template<typename T>
bool ShmBuffer<T>::write(const T& value)
{
  boost::interprocess::scoped_lock<ShmMutex> lock(m_mutex);  
  ShmDeque *buff = m_buffOffset.get();

  // 共有メモリセグメントは有限なのでバッファサイズを制限する
  if(buff->size() == m_maxBufferSize){ 
    buff->pop_front();
  }
  buff->push_back(value);
  
  return true;
}

////////////////////////////////////////////////////////////////////////////////
/*!
 * @brief バッファ先頭の要素を読み込む。
 *  バッファ先頭の要素をread読み込む。要素はバッファから取り除かれる。
 *
 */
template<typename T>
bool ShmBuffer<T>::read(T& readVal)
{
  boost::interprocess::scoped_lock<ShmMutex> lock(m_mutex);
  ShmDeque *buff = m_buffOffset.get();

  if(buff->empty()){
    return false;
  }else{
    readVal = buff->front();
    buff->pop_front();
    return true;
  }
}

////////////////////////////////////////////////////////////////////////////////
/*!
 * @brief バッファ先頭の要素への参照を返す
 * バッファ先頭の要素への参照を返す。先頭の要素はバッファに残る。
 */
template<typename T>
T& ShmBuffer<T>::front()
{
  boost::interprocess::scoped_lock<ShmMutex> lock(m_mutex);
  return m_buffOffset.get()->front();
}


////////////////////////////////////////////////////////////////////////////////
/*!
 * @brief バッファが空か否か確認する。
 * @return 空であれば true、そうでなければ false
 */
template<typename T>
bool ShmBuffer<T>::isEmpty() 
{
  boost::interprocess::scoped_lock<ShmMutex> lock(m_mutex);
  return m_buffOffset.get()->empty();
}
