/*!
 * @brief プリミティブ型用のShmInPort<T>の実装
 * プリミティブ型用のShmInPort<T>の実装。
 * プリミティブ型でなくても boost::interprocess の
 * コンテナと同様に 共有メモリ利用を前提に設計されている
 * クラスであればこの実装が適用できる。
 */

#ifndef SHM_IN_PORT_PRIMITIVE_IMPL_HPP
#define SHM_IN_PORT_PRIMITIVE_IMPL_HPP


#include <unistd.h>
#include <iostream>
#include <rtm/CorbaPort.h>
#include <shmport/ShmInPort_primitive.hpp>
#include <shmport/ShmPortCommon.hpp>
#include <shmport/ShmBuffer.hpp>


//#define USE_ON_CONNECTED
//#define USE_PRE_ON_ACTIVATED
#define USE_INNER_LISTENER // 下記リスナが使えるのは OpenRTm-aist 1.1.0 以降?

// open()他の隠蔽のため、privateメンバにアクセスできる内部クラスのリスナを使う
template<typename T>
class ShmInPort<T>::ActivationListener : public RTC::PreComponentActionListener
{
public:
  ActivationListener(ShmInPort<T>& port) :
    m_port(port)
  { }
  virtual void operator()(RTC::UniqueId execContextID)
  { m_port.open();  }
private:
  ActivationListener(); // 意図しない初期化を回避するために privateにする
  ShmInPort<T>& m_port;
};

template<typename T>
class ShmInPort<T>::DeactivationListener : public RTC::PostComponentActionListener
{
public:
  DeactivationListener(ShmInPort<T>& port) :
    m_port(port)
  { }
  
  virtual void operator()(RTC::UniqueId execContextID, RTC::ReturnCode_t retCode)
  { m_port.closeShmSegment();  }
private:
  DeactivationListener(); // 意図しない初期化を回避するために privateにする
  ShmInPort<T>& m_port;
};


namespace // 下記シンボルを隠すため、匿名名前空間を使う
{
class MsgPrinter
{
public:
  MsgPrinter(const std::string& msg)
  {
    using namespace std;
    cout << msg << endl;
  }
};
};


////////////////////////////////////////////////////////////////////////////////
template<typename T>
ShmInPort<T>::ShmInPort(const char *portName, T& boundVar) :
  m_name(portName),
  m_shmSegment(NULL),
  m_shmBuffer(NULL),
  m_servicePort(portName),
  m_boundVar(boundVar)
{
  
}

////////////////////////////////////////////////////////////////////////////////
template<typename T>
ShmInPort<T>::~ShmInPort()
{}


////////////////////////////////////////////////////////////////////////////////
/***/
template<typename T>
bool ShmInPort<T>::openShmSegment(const char *segmentID)
{
  using namespace std;
  using namespace boost::interprocess;

  if(m_shmSegment != NULL){
    //closeShmSegment();
    return true; // ? To Do: 接続を切断して別のRTCに接続する場合への対応 ?
  }
  try{
    cout << "Opening a shared memory segment named " << segmentID << endl;
    m_shmSegment = new managed_shared_memory(open_only, segmentID);
  }catch(interprocess_exception &e){
    cerr << "Boost.Interprocess exception: " << e.what() << endl;
    return false;
  }

  if(m_shmSegment != NULL){
    return true;
  }else{
    return false;
  }
}


////////////////////////////////////////////////////////////////////////////////
template<typename T>
bool ShmInPort<T>::findInstance(const char *instanceID)
{
  using namespace std;
  using namespace boost::interprocess;  
  std::pair<ShmBuffer<T>*, std::size_t> res;

  if(m_shmSegment == NULL){
    return false;
  }

  try{
    cout << "Searching for ShmBuffer in shmSegment" << endl;
    res = m_shmSegment->find< ShmBuffer<T> >(instanceID);
    if(res.first == NULL){
      cerr << "Instance `" << instanceID
	   << "' not found in the shared memory" << endl;
    }else{
      cerr << "found" << endl;
    }
  }catch(interprocess_exception &e){
    cerr << "Boost.Interprocess exception: " << e.what() << endl;
    return false;
  }

  m_shmBuffer = res.first;
  return true;
}

////////////////////////////////////////////////////////////////////////////////
template<typename T>
void ShmInPort<T>::closeShmSegment()
{
  using namespace std;
  cerr << "closing shmSegment..." << endl;
  
  delete m_shmSegment;
  m_shmSegment = NULL;
  m_shmBuffer = NULL;
}

////////////////////////////////////////////////////////////////////////////////
/**
 *  
 */
template<typename T>
void ShmInPort<T>::open()
{
  // 共有データへのアクセスのため、共有メモリを探索する
  using namespace std;
  bool succeeded = false;
  bool completed = false;


  while(!completed){
    try{
      cerr << "Trying to get shmSegmentID" << endl;
      succeeded = openShmSegment(m_consumer->shmSegmentID());
      if(succeeded){
	cout << "shmSegment is open." << endl;
	succeeded = findInstance(m_consumer->shmInstanceID());
      }

      if(succeeded){
#if 0	
	cout << "an instance has been found in shmSegment\n"
	     << "  instance ID: "  << m_consumer->shmInstanceID() << "\n"
	     << "  address: " << hex << m_shmBuffer  << "\n"
	     << "  value ==  " << dec << m_shmBuffer->value() << endl;
#endif	
	completed = true;
      }
    }catch(CORBA::SystemException &e){
      const long DURATION = 10*000;
      cerr << "CORBA SystemException: \n"
	   << "  " << e._name() << "\n"
	   << "  " << e._rep_id() << "\n"
	   << endl;
      cerr << "Failed to get shmSegmentID.\n Retrying..." << endl;
      coil::usleep(DURATION);
    }
  }

}

////////////////////////////////////////////////////////////////////////////////
template<typename T>
void ShmInPort<T>::attachTo(RTC::DataFlowComponentBase& rtc)
{
#ifdef USE_INNER_LISTENER
  static ShmInPort<T>::ActivationListener  actListener(*this);
  static ShmInPort<T>::DeactivationListener deactListener(*this);
#endif
  
  m_servicePort.registerConsumer(ShmPort::SVC_INSTANCE_NAME, ShmPort::SVC_TYPE_NAME, m_consumer);
  rtc.addPort(m_servicePort);

  
#ifdef USE_INNER_LISTENER
  rtc.addPreComponentActionListener(RTC::PRE_ON_ACTIVATED,
				    &actListener,
				    false); // autoclean無効化
  
  rtc.addPostComponentActionListener(RTC::POST_ON_DEACTIVATED,
				     &deactListener,
				     false); // autoclean無効化  
#endif
}



////////////////////////////////////////////////////////////////////////////////
template<typename T>
bool ShmInPort<T>::read()
{
  bool succeeded = false;
  if(m_shmBuffer == NULL){
    return false;
  }

  succeeded = m_shmBuffer->read(m_boundVar);
  return succeeded;
}

////////////////////////////////////////////////////////////////////////////////
template<typename T>
bool ShmInPort<T>::peek(T& retVal)
{
  if(m_shmBuffer == NULL){
    return false;
  }

  if(m_shmBuffer->isEmpty()){
    return false;
  }
  
  retVal = m_shmBuffer->front();
  return true;
}

////////////////////////////////////////////////////////////////////////////////
/**
 *
 * @see RTC::InPort::isNew
 */
template<typename T>
bool ShmInPort<T>::isNew() 
{
  if(m_shmBuffer == NULL){ 
    return false;
  }
  
  return !m_shmBuffer->isEmpty();
}

#endif // SHM_IN_PORT_PRIMITIVE_IMPL_HPP
