/*!
 * @brief プリミティブ型用のShmOutPort<T>の実装
 * 
 * プリミティブ型用のShmOutPort<T>の実装
 * プリミティブ型でなくても boost::interprocess の
 * コンテナと同様に 共有メモリ利用を前提に設計されている
 * クラスであればこの実装が適用できる。 
 */
#ifndef SHM_OUT_PORT_PRIMITIVE_IMPL_HPP
#define SHM_OUT_PORT_PRIMITIVE_IMPL_HPP


#include <boost/uuid/uuid.hpp>
#include <boost/uuid/uuid_generators.hpp>
#include <boost/uuid/uuid_io.hpp>
#include <boost/lexical_cast.hpp>
#include <rtm/CorbaPort.h>
#include <shmport/ShmOutPort_primitive.hpp>
#include <shmport/ShmBuffer.hpp>
#include <shmport/ShmPortCommon.hpp>


////////////////////////////////////////////////////////////////////////////////
namespace // 以下のシンボルの隠蔽のため匿名名前空間を使う
{ 
  const size_t MAX_BUFFER_SIZE = 256;
  
  //* prefix + UUID の IDを生成する
  class IDGen
  {
  public:
    IDGen(const char* prefix) :
      m_prefix(prefix)
    {}
    
    const char* c_str()
    {
      m_id = m_prefix;
      m_id += boost::lexical_cast<std::string>(m_gen());
      return m_id.c_str();
    }
  
  private:
    boost::uuids::basic_random_generator<boost::mt19937> m_gen;
    std::string m_prefix;
    std::string m_id;
  };
};


template<typename T> const size_t ShmOutPort<T>::DEFAULT_SEGMENT_SIZE = 16384;


////////////////////////////////////////////////////////////////////////////////
/**
 *
 */
template<typename T>
ShmOutPort<T>::ShmOutPort(const char *portName,
			  size_t segmentSize) :
  m_name(portName),
  m_shmSegmentID(IDGen("ShmSeg").c_str()),
  m_shmInstanceID(IDGen("ShmBuffer").c_str()),
  m_shmSegment(boost::interprocess::create_only, m_shmSegmentID.c_str(), segmentSize),
  m_servicePort(m_name.c_str())
{
  using namespace std;
  //IDGen instanceID("ShmBuffer");
  cout << "Constructing ShmBuffer<T> in segment " << m_shmSegmentID << endl;
  cout << " instance ID: " << m_shmInstanceID << endl;
  
  m_shmBuffer =
    m_shmSegment.construct< ShmBuffer<T> >(m_shmInstanceID.c_str())(&m_shmSegment, MAX_BUFFER_SIZE);

  if(m_shmBuffer == NULL){
    cerr << "Failed to construct ShmBuffer<T> in segment" << endl;
  }
  //m_provider.setShmInstanceID(instanceID.c_str());
  //m_provider.setShmInstanceID(m_shmSegment.get_instance_name< ShmBuffer<T> >(m_shmBuffer));
  //m_provider.setShmInstanceID(m_shmSegment.get_instance_name(m_shmBuffer));

#if 1  
  m_provider.setShmSegmentID(m_shmSegmentID.c_str());
  m_provider.setShmInstanceID(m_shmInstanceID.c_str());
#endif  
}


////////////////////////////////////////////////////////////////////////////////
/**
 */
template<typename T>
ShmOutPort<T>::~ShmOutPort()
{
  m_shmSegment.destroy_ptr(m_shmBuffer);
  boost::interprocess::shared_memory_object::remove(m_shmSegmentID.c_str());
}


////////////////////////////////////////////////////////////////////////////////
/**
 *  @param rtc ポートを追加するRTC
 */
template<typename T>
void ShmOutPort<T>::attachTo(RTC::DataFlowComponentBase& rtc)
{
  //m_servicePort.registerProvider(m_name.c_str(), ShmPort::PORT_TYPE_NAME, m_provider);
  m_servicePort.registerProvider(ShmPort::SVC_INSTANCE_NAME, ShmPort::SVC_TYPE_NAME, m_provider);
  rtc.addPort(m_servicePort);
}


////////////////////////////////////////////////////////////////////////////////
/**
 *
 */
template<typename T>
void ShmOutPort<T>::write(const T& value)
{
  m_shmBuffer->write(value);
}


#endif // SHM_OUT_PORT_PRIMITIVE_IMPL_HPP
